import logging
import os
from modelbest_sdk.file_format.mbtable import MbTable
from modelbest_sdk.file_format.mbtable_builder import MbTableBuilder


class Utils:
    @staticmethod
    def load_from_file(path):
        logging.info('Read DatasetContext from file %s', path)
        mbtable = MbTable(path)
        if mbtable.get_entry_count() < 1:
            raise ValueError('Empty file, please double check the content.')
        '''
        Always read the latest record. Could support version lookup in the future.
        '''
        record = mbtable.read(0)
        return record
    
    @staticmethod
    def save_to_file(path, bin):
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        builder = MbTableBuilder(path)
        builder.write(bin)
        builder.flush()
        logging.info('Write DatasetInfoList to file %s', path)


from collections import defaultdict
import numpy as np
import torch

from modelbest_sdk.dataset.collater.collater import Collater
from modelbest_sdk.dataset.sampler.weighted_dataset import WeightedDataset
from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class BatchedDataset(torch.utils.data.IterableDataset):
    def __init__(self, context, weighted_dataset, batch_size, max_len, drop_last=False):
        self.context = context
        self.weighted_dataset = weighted_dataset
        self.max_total_length = batch_size * max_len
        self.batch_size = 1
        self.drop_last = drop_last
        
        
        self.buffer = []
        self.current_length = 0
        

    def put(self, data):
        self.buffer.append(data)
        self.current_length += len(data['doc'].token_ids)
    
    def pop(self):
        lengths = []
        indexes = defaultdict()
        
        inputs = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)
        targets = torch.full((self.batch_size, self.max_total_length), dtype=torch.int32, fill_value=-100)
        dataset_ids = torch.full((self.batch_size, self.max_total_length), dtype=torch.int32, fill_value=-1)
        position_ids = torch.zeros((self.batch_size, self.max_total_length), dtype=torch.int32)

        span_begin = 0
        while self.buffer:
            data = self.buffer.pop()
            chunk = data['chunk']
            index = data['index']
            dataset_idx = data['dataset_idx']
            doc: BaseDoc = data['doc']
            token_ids = doc.token_ids
            mask = doc.mask
            tag = doc.tag
            target_ids = [token_id if not mask_val else -100 for token_id, mask_val in zip(token_ids[1:], mask[1:])] + [-100]
            
            span_end = span_begin + len(token_ids)
            # TODO: what if the inputs is longer than max_total_length?
            # RuntimeError: The expanded size of the tensor (16) must match the existing size (385) at non-singleton dimension 0.  Target sizes: [16].  Tensor sizes: [385]
            inputs[0, span_begin:span_end] = torch.tensor(token_ids, dtype=torch.int32)
            targets[0, span_begin:span_end] = torch.tensor(target_ids, dtype=torch.int32)
            dataset_ids[0, span_begin:span_end] = torch.tensor(dataset_idx, dtype=torch.int32)
            position_ids[0, span_begin:span_end] = torch.from_numpy(np.arange(len(token_ids), dtype=np.int32))
            lengths.append(len(token_ids))
            if indexes.get(int(dataset_idx)) is None:
                indexes[int(dataset_idx)] = {}
            if indexes[int(dataset_idx)].get(chunk) is None:
                indexes[int(dataset_idx)][chunk] = []
                # TODO: better way
            indexes[int(dataset_idx)][chunk].append(index)
            span_begin = span_end
        cu_seqlens = torch.cat(
            [torch.tensor([0] + lengths).cumsum(dim=-1), torch.tensor([self.max_total_length], dtype=torch.int32)],
            dim=0,
        ).int()
        batch = {
            "inputs": inputs,
            "targets": targets,
            "dataset_ids": dataset_ids,
            "indexes": indexes,
            "cu_seqlens": cu_seqlens,
            "max_seqlen": int(torch.max(cu_seqlens[1:] - cu_seqlens[:-1])),
            "lengths": torch.tensor(sum(lengths)).int(),
            "position_ids": position_ids,
            # TODO: add tags
        }
        print(batch['inputs'].shape)
        self.current_length = 0
        return batch
        # TODO: return data directly, and pack in collate_fn
        

    def will_exceed(self, data):
        return self.current_length + len(data['doc'].token_ids) > self.max_total_length
        

    def __iter__(self):
        for data in self.weighted_dataset:
            if self.will_exceed(data):
                yield self.pop()
            self.put(data)
        # TODO: drop last batch if it is not full

if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)

    dataset_info_list = [
        DatasetInfo(
            path="/home/emr-user/modelbest_sdk/test/base_doc_simple",
            weight=1,
            max_epoch=1
        ),
        DatasetInfo(
            path="/home/emr-user/modelbest_sdk/test/base_doc_easy",
            weight=2,
            max_epoch=2
        )
    ]
    
    dataset_info_list = DatasetInfoList(dataset_info_list)
    
    w_dataset = WeightedDataset(
        context=context,
        dataset_info_list=dataset_info_list
    )
    
    dataset = BatchedDataset(
        context=context,
        weighted_dataset=w_dataset,
        batch_size=16,
        max_len=1
    )
    
    for data in dataset:
        print(data)
        print(data['indexes'].keys())
        print(data['indexes'].values())
    


import torch
from modelbest_sdk.dataset.collater.batched_dataset import BatchedDataset
from modelbest_sdk.dataset.sampler.weighted_dataset import WeightedDataset
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import DatasetInfo, DatasetInfoList
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class TableRecordDataset:
    def __init__(
        self,
        context: DatasetContext,
        batch_size=1,
        max_len=4096,
        prefetch_chunk_cnt=20,
        chunk_size=1024,
        num_workers=1,
        prefetch_factor=20,
        cuda_prefetch=True
    ):
        dataset_info_list = DatasetInfoList.load_from_file(context.dataset_config_path)
        
        self.weighted_dataset = WeightedDataset(
            context=context, 
            dataset_info_list=dataset_info_list,
            prefetch_chunk_cnt=prefetch_chunk_cnt,
            chunk_size=chunk_size
        )
        
        self.batched_dataset = BatchedDataset(
            context=context,
            weighted_dataset=self.weighted_dataset,
            batch_size=batch_size,  
            max_len=max_len
        )
        
        self.dataloader = torch.utils.data.DataLoader(
            dataset=self.batched_dataset, 
            collate_fn=default_collate, 
            num_workers=1,
            prefetch_factor=prefetch_factor
        )
        
        self.cuda_prefetch = cuda_prefetch
        
        if self.cuda_prefetch:
            from modelbest_sdk.dataset.cuda_prefetcher import CudaPrefetcher
            self.cuda_prefetcher = CudaPrefetcher(
                context,
                self.dataloader
            )
        
    
    def __iter__(self):
        if not self.cuda_prefetch:
            for batch in self.dataloader:
                yield batch
        else:
            for batch in self.cuda_prefetcher:
                yield batch
    
    def checkpoint(self):
        return self.weighted_dataset.checkpoint()
    
    def load_checkpoint(self, checkpoint):
        self.weighted_dataset.load_checkpoint(checkpoint)
        
    def update(self, dataset_entries):
        self.weighted_dataset.update(dataset_entries)
        
def default_collate(batch):
    return batch[0]


import os
import sys
import thriftpy2

cur_path = os.path.split(os.path.realpath(__file__))[0]

from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.file_format.mbtable_builder import MbTableBuilder
from thriftpy2.utils import serialize


dc_thrift = thriftpy2.load(os.path.join(cur_path,"../proto/dataset_context.thrift"), module_name="context_thrift")

def test_read_from_file():
    test_path = os.path.join('/tmp', 'test_context.sstable')
    table = MbTableBuilder(test_path)

    nc = dc_thrift.DatasetContext(
        rank=1,
        world_size=10,
        dataset_config_path="/path/to/datainfo_list",
        dataset_checkpoint_path="/path/to/checkpoint_list",
    )
    table.write(serialize(nc))
    table.flush()
    print('write data to file %s', test_path)

    read_nc = DatasetContext(test_path).context
    print(read_nc)
    assert read_nc.rank == nc.rank
    assert read_nc.world_size == nc.world_size
    assert read_nc.dataset_config_path == nc.dataset_config_path
    assert read_nc.dataset_checkpoint_path == nc.dataset_checkpoint_path


if __name__ == "__main__":
  test_read_from_file()



import os

from modelbest_sdk.file_format.mbtable import MbTable, MbTableIterator

'''
MbTablePartition is designed to manage a directory of mbtable files, and provide a way to iterate over the rows of the files.
Note that each file can contain max 2 << 30 records.
'''
class MbTablePartition:
    def __init__(self, partition_path: str):
        """
        calculate total count of rows in mbtable partition
        """
        self.partition_path = partition_path
        self.total_count = 0
        self.file_name_list = [file for file in os.listdir(partition_path)]
        self.file_name_list.sort()
        self.abs_path_list = [os.path.join(partition_path, file) for file in self.file_name_list]
        self.abs_path_index_dict = {file: i for i, file in enumerate(self.abs_path_list)}
        self.file_handle_list = [MbTable(file) for file in self.abs_path_list]
        self.file_row_counts = []
        self.cumulative_row_counts = []

        cumulative_count = 0
        for handle in self.file_handle_list:
            count = handle.get_entry_count()
            self.total_count += count
            self.file_row_counts.append(count)
            
            cumulative_count += count
            self.cumulative_row_counts.append(cumulative_count)

    def get_file_index_and_position(self, n):
        """
        根据行号n找到对应的文件以及该行在文件中的位置
        """
        if n > self.total_count:
            return None, None  # n 超出总行数

        # 确定n在哪个文件中
        for i, cumulative in enumerate(self.cumulative_row_counts):
            if n <= cumulative:
                # 确定在当前文件中的具体位置
                position_in_file = n - (self.cumulative_row_counts[i-1] if i > 0 else 0)
                return i, position_in_file
        
        return None, None  # 如果没有找到，返回None
    
    def get_total_count(self) -> int:
        return self.total_count
    
    def get_file_path(self, file_index):
        return self.abs_path_list[file_index]

    def get_next_file_index(self, file_index):
        if file_index == len(self.abs_path_list) - 1:
            return None
        else:
            return file_index + 1
        
    
class MbTablePartitionIterator:
    def __init__(self, mbtable_partition: MbTablePartition, start_index=0, max_iterations=None):
        self.mbtable_partition = mbtable_partition
        self.max_iterations = max_iterations
        self.iterations_count = 0
        assert start_index >= 0 and start_index <= mbtable_partition.get_total_count()
        self.current_file_index, self.current_pos = mbtable_partition.get_file_index_and_position(start_index)
        self.iterator = None
        
    def __iter__(self):
        return self
    
    def __enter__(self):
        if self.iterator is None:
            self.iterator = MbTableIterator(self.mbtable_partition.get_file_path(self.current_file_index), self.current_pos, self.max_iterations)
            self.iterator.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.iterator.__exit__(exc_type, exc_val, exc_tb)
    
    def __next__(self):
        if self.max_iterations is not None and self.iterations_count >= self.max_iterations:
            raise StopIteration

        # 检查是否需要初始化或重置迭代器
        if self.iterator is None:
            # 计算剩余的迭代次数，如果 max_iterations 未定义，则不限制迭代器
            remaining_iterations = None if self.max_iterations is None else self.max_iterations - self.iterations_count
            self.iterator = MbTableIterator(self.mbtable_partition.get_file_path(self.current_file_index), self.current_pos, remaining_iterations)
            self.iterator.__enter__()

        try:
            record = self.iterator.__next__()
            self.iterations_count += 1  # 更新已迭代的数量
            return record
        except StopIteration:
            self.current_file_index = self.mbtable_partition.get_next_file_index(self.current_file_index)
            if self.current_file_index is None:
                raise StopIteration
            self.current_pos = 0
            self.iterator = None  # 重置迭代器以便于下次调用时重新初始化
            return self.__next__()


if __name__ == '__main__':
    mbtable_partition = MbTablePartition('test/partition_data')
    total_count = mbtable_partition.get_total_count()
    print(f"total count: {total_count}")
    with MbTablePartitionIterator(mbtable_partition, start_index=9, max_iterations=11) as iterator:
        for row in iterator:
            print(f"row: {row}")


import copy
import json
from typing import Iterable
import numpy as np
import torch

from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


class CudaPrefetcher(Iterable):
    """
    Wrap around a batch iterator for asynchornously copying data to gpu to shield memcpy latency.
    """

    def __init__(self, context: DatasetContext, loader):
        self.context = context
        self.loader = iter(loader)
        self.tp_size = context.tp_size
        self.tp_rank = context.tp_rank
        self.stream = torch.cuda.Stream()
        self.preload()

    def preload(self):
        try:
            if self.tp_size > 1:
                if self.tp_rank == 0:
                    data = next(self.loader)
                    print(f"Rank {self.context.rank}, Preload data done.")
                    d = {}
                    with open(f"/dev/shm/TP_{self.context.tp_rank}.bin", "wb") as fb:
                        for key in data.keys():
                            if isinstance(data[key], torch.Tensor):
                                np_cur_data = data[key].cpu().numpy()
                                bs = np_cur_data.tobytes()
                                fb.write(bs)
                                d[key] = ["TORCH", str(np_cur_data.dtype), len(bs)] + list(np_cur_data.shape)
                            else:
                                d[key] = data[key]
                        try:
                            _ = json.dumps(d)
                        except TypeError:
                            print(d)
                        with open(f"/dev/shm/TP_{self.context.tp_rank}.json", "w") as f:
                            json.dump(d, f)
                torch.cuda.synchronize()
                if self.tp_rank != 0:
                    with open(f"/dev/shm/TP_{self.context.tp_rank}.json", "r") as f:
                        data = json.load(f)
                    with open(f"/dev/shm/TP_{self.context.tp_rank}.bin", "rb") as fb:
                        bs = fb.read()
                        offset = 0
                        for key in data.keys():
                            if isinstance(data[key], list) and len(data[key]) > 1 and data[key][0] == "TORCH":
                                nw_offset = offset + data[key][2]
                                data[key] = torch.from_numpy(
                                    np.frombuffer(bs[offset:nw_offset], dtype=data[key][1])
                                    .reshape(data[key][3:])
                                    .copy()
                                )
                                offset = nw_offset
                self.data = data
            else:
                self.data = next(self.loader)
        except StopIteration:
            self.data = None
            return
        with torch.cuda.stream(self.stream):
            for key in self.data.keys():
                if isinstance(self.data[key], torch.Tensor):
                    self.data[key] = self.data[key].cuda(non_blocking=True)

    def __next__(self):
        torch.cuda.current_stream().wait_stream(self.stream)
        data = copy.deepcopy(self.data)
        self.preload()
        return data

    def __iter__(self):
        return self



import sys
import numpy as np
import torch
from modelbest_sdk.dataset.mbtable_iterable_dataset import MBTableIterableDataset
from modelbest_sdk.dataset.sampler.sampler import Sampler
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext

MAX_EPOCH = 2**31 - 1

class WeightedDataset(Sampler, torch.utils.data.IterableDataset):
    def __init__(
        self, 
        context: DatasetContext, 
        dataset_info_list: DatasetInfoList,
        prefetch_chunk_cnt=20,
        chunk_size=1024
        ):
        
        self.context = context
        self.dataset_info_list = dataset_info_list.dataset_info_list

        self.dataset_weights = []
        self.datasets: List[MBTableIterableDataset] = []
        self.datasets_iter = []
        weights = []
        for dataset_info in self.dataset_info_list:
            path = dataset_info.path
            weight = dataset_info.weight
            if weight == 0:
                print(f"Dataset {path} has weight 0, skip it")
                continue
            max_epoch = dataset_info.max_epoch if dataset_info.max_epoch is not None else MAX_EPOCH
            dataset = MBTableIterableDataset(
                context=context,
                path=path,
                max_epoch=max_epoch,
                prefetch_chunk_cnt=prefetch_chunk_cnt,
                chunk_size=chunk_size
            )
            self.datasets.append(dataset)
            self.datasets_iter.append(iter(dataset))
            weights.append(weight)
        weights = np.array(weights)
        self.dataset_weights = weights / weights.sum()
        self.remain = len(self.datasets)
        
            
    def sample(self) -> int:
        choice = np.random.choice(len(self.dataset_weights), p=self.dataset_weights)
        return choice
    
    def __iter__(self):
        while True:
            if all(d.exhausted for d in self.datasets):
                print(f"All dataset exhaust on rank {self.context.rank}")
                break
            idx = self.sample()
            if self.datasets[idx].exhausted:
                print(f"Dataset {idx} exhaust on rank {self.context.rank}")
                continue
            chosen_iter = self.datasets_iter[idx]
            try:
                data = next(chosen_iter)
                if data is None:
                    continue
                data['dataset_idx'] = idx
                yield data
            except StopIteration:
                continue


    def set_seed():
        # TODO: implement set_seed
        pass
    
    def checkpoint(self):
        checkpoint_list = [dataset.checkpoint() for dataset in self.datasets]
        return DatasetCheckpointList(
            checkpoint_list=checkpoint_list,
            world_size=self.context.world_size,
            tp_size=self.context.tp_size
        )
    
    def load_checkpoint(self, dataset_checkpoint_list: DatasetCheckpointList):
        for i, checkpoint in enumerate(dataset_checkpoint_list.checkpoint_list):
            self.datasets[i].load_checkpoint(checkpoint)
            
    def update(self, dataset_entries: Dict[int, Dict]):
        for i, entries in dataset_entries.items():
            self.datasets[i].update(entries)
    

if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)

    dataset_info_list = [
        DatasetInfo(
            path="/home/emr-user/modelbest_sdk/test/base_doc_simple",
            weight=1,
            max_epoch=1
        ),
        DatasetInfo(
            path="/home/emr-user/modelbest_sdk/test/base_doc_easy",
            weight=2,
            max_epoch=2
        )
    ]
    
    dataset_info_list = DatasetInfoList(dataset_info_list=dataset_info_list)
    
    dataset = WeightedDataset(
        context=context,
        dataset_info_list=dataset_info_list
    )

    for data in dataset:
        print(data)


import thriftpy2
from modelbest_sdk.file_format.mbtable_builder import MbTableBuilder
from modelbest_sdk.file_format.mbtable_partition import MbTablePartition, MbTablePartitionIterator
from thriftpy2.utils import serialize, deserialize


# for i in range(10):
#     builder = MbTableBuilder(f"test/partition_data/part_{i}.mbt")
#     for j in range(100):
#         builder.write(f"file_{i}_value_{j}")
#     builder.add_metadata("meta_key", "meta_value")
#     builder.flush()
# builder = MbTableBuilder(f"test/base_doc_simple/part_1.mbt")
# dataset = MbTablePartitionIterator(MbTablePartition("test/base_doc"))
# for i, data in enumerate(dataset):
#     builder.write(data)
# builder.flush()
base_doc_thrift = thriftpy2.load("modelbest_sdk/proto/traindoc.thrift", module_name="proto_base_doc_thrift")

for i in range(2):
    builder = MbTableBuilder(f"test/base_doc_easy/part_{i}.mbt")
    for j in range(10):
        doc = base_doc_thrift.BaseDoc()
        doc.docid = f"doc_{i}_{j}"
        doc.token_ids = [-1, 2, 3, 4, 5, 6, 0]
        doc.mask = [False, False, False, False, False, False, True]
        doc.tag=["mock"]
        doc_str = serialize(doc)
        builder.write(doc_str)
    builder.flush()


import logging
import os
import thriftpy2
from thriftpy2.utils import deserialize, serialize

from modelbest_sdk.dataset.thrift_wrapper.utils import Utils

proto_dir = os.path.join(os.path.dirname(__file__), "../..", "proto")

dc_thrift = thriftpy2.load(os.path.join(proto_dir, "dataset_context.thrift"), module_name="context_thrift")
logger = logging.getLogger(__name__)


class DatasetContext:
    def __init__(self, rank=0, world_size=1, tp_size=1, tp_rank=0, num_workers=1, dataset_config_path=None, dataset_checkpoint_path=None):
        self.rank = rank
        self.world_size = world_size
        self.tp_size = tp_size
        self.tp_rank = tp_rank
        self.num_workers = num_workers
        self.dataset_config_path = dataset_config_path
        self.dataset_checkpoint_path = dataset_checkpoint_path

    @staticmethod
    def load_from_file(path):
        return DatasetContext.from_thrift(deserialize(dc_thrift.DatasetContext(), Utils.load_from_file(path)))

    def save_to_file(self, path):
        Utils.save_to_file(path, serialize(self.to_thrift()))

    def to_thrift(self):
        return dc_thrift.DatasetContext(
            rank=self.rank,
            world_size=self.world_size,
            tp_size=self.tp_size,
            tp_rank=self.tp_rank,
            num_workers=self.num_workers,
            dataset_config_path=self.dataset_config_path,
            dataset_checkpoint_path=self.dataset_checkpoint_path
        )

    @staticmethod
    def from_thrift(thrift_context):
        return DatasetContext(
            rank=thrift_context.rank,
            world_size=thrift_context.world_size,
            tp_size=thrift_context.tp_size,
            tp_rank=thrift_context.tp_rank,
            num_workers=thrift_context.num_workers,
            dataset_config_path=thrift_context.dataset_config_path,
            dataset_checkpoint_path=thrift_context.dataset_checkpoint_path
        )
    
    def __repr__(self) -> str:
        return f"DatasetContext(rank={self.rank}, world_size={self.world_size}, tp_size={self.tp_size}, tp_rank={self.tp_rank}, num_workers={self.num_workers}, dataset_config_path={self.dataset_config_path}, dataset_checkpoint_path={self.dataset_checkpoint_path})"




import queue
import threading
from typing import List

from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import Chunk


class CacheNotFilled(Exception):
    pass


class CacheFull(Exception):
    pass


class AllCached(Exception):
    pass


class PrefetchCache(object):
    """
    Primitive for concurrently performing tasks and block when cache is full.
    Manage shared states to avoid racing, ensuring at most self.size tasks are cached
    Typical use case
    1. submit tasks to be cached
    cache.submit(tasks)
    2. within producer threads: get task, do some process, and put the results back
    while True:
        try:
            task = cache.pull_task()
            result = do_something(task)  # heavy work
            cache.put_result(result)
        except CacheFull:
            time.sleep(0.2)
        except AllCached:
            break
    3. get results within the consumer thread
    # CacheNotFilled: tasks not exhausted but no task are currently pulled, supressed by block=True
    # StopIteration: tasks are exhausted and all results are consumed
    result = cache.get(block=False)
    do_something(result)
    """

    def __init__(self, size):
        self.size = size
        self._lock = threading.RLock()
        self._count = self._finished = self._queue = self._tasks = None

    def submit(self, tasks):
        with self._lock:
            self._queue = queue.Queue()  # cache for the task results
            self._tasks = tasks  # iterator of tasks to be performed
            self._count = 0  # counting the expected results, manually maintaining count to avoid racing
            self._finished = False

    def pull_task(self) -> Chunk:
        with self._lock:
            if self._count + 1 > self.size:
                raise CacheFull
            try:
                task = next(self._tasks)
                self._count += 1
                return task
            except StopIteration:
                self._finished = True
                raise AllCached

    @property
    def finished(self):
        with self._lock:
            return self._finished

    @property
    def count(self):
        with self._lock:
            return self._count

    def put_result(self, result):
        self._queue.put(result)

    def get(self, block=False):
        with self._lock:
            if self._count <= 0:
                if self._finished:
                    raise StopIteration
                elif not block:
                    raise CacheNotFilled
            self._count -= 1
        data = self._queue.get(block=True)
        return data




import math
import random


def random_range(start, stop=None, step=None):
    """
    Generator of non-repeated random permutation with the same inteface of python
    `range`. Obtained from https://stackoverflow.com/a/53551417
    The random.shuffle(list) and random.sample(list, len(list)) require
    materialize the lists, which result in a long initalization period.
    """
    if stop is None:
        start, stop = 0, start
    if step is None:
        step = 1
    # Use a mapping to convert a standard range into the desired range.
    mapping = lambda i: (i * step) + start
    # Compute the number of numbers in this range.
    maximum = int(math.ceil((stop - start) / step))
    if maximum == 0:
        # early return with empty range
        yield from ()
        return
    # Seed range with a random integer.
    value = random.randint(0, maximum)
    # Construct an offset, multiplier, and modulus for a linear
    # congruential generator. These generators are cyclic and
    # non-repeating when they maintain the properties:
    #
    #   1) "modulus" and "offset" are relatively prime.
    #   2) ["multiplier" - 1] is divisible by all prime factors of "modulus".
    #   3) ["multiplier" - 1] is divisible by 4 if "modulus" is divisible by 4.

    # Pick a random odd-valued offset.
    offset = random.randint(0, maximum) * 2 + 1
    # Pick a multiplier 1 greater than a multiple of 4.
    multiplier = 4 * (maximum // 4) + 1
    # Pick a modulus just big enough to generate all numbers (power of 2).
    modulus = int(2 ** math.ceil(math.log2(maximum)))
    # Track how many random numbers have been returned.
    found = 0
    while found < maximum:
        # If this is a valid value, yield it in generator fashion.
        if value < maximum:
            found += 1
            yield mapping(value)
        # Calculate the next value in the sequence.
        value = (value * multiplier + offset) % modulus


class Range(object):
    def __init__(self, start, stop, step):
        self.start = start
        self.stop = stop
        self.step = step

    def __repr__(self):
        return f"Range({self.start}, {self.stop}, {self.step})"

    def iterate(self):
        yield from range(self.start, self.stop, self.step)

    def list(self):
        return list(range(self.start, self.stop, self.step))

    def subrange(self, split, nsplits):
        # strided spliting range params
        # e.g., [0, 3, 5, 7, 9] can be split into [0, 5, 9] and [3, 7]
        return Range(self.start + self.step * split, self.stop, self.step * nsplits)

    def random_iterate(self):
        yield from random_range(self.start, self.stop, self.step)



import torch

from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext


def get_worker_info():
    worker_info = torch.utils.data.get_worker_info()
    if worker_info is None:
        num_workers, worker_id = 1, 0
    else:
        num_workers, worker_id = worker_info.num_workers, worker_info.id
    return num_workers, worker_id

def get_context():
    context = DatasetContext(world_size=1, rank=0, num_workers=1)
    return context


import ctypes
import os

class ByteArray(ctypes.Structure):
    _fields_ = [("data", ctypes.POINTER(ctypes.c_char)),
                ("length", ctypes.c_size_t)]

lib_path = os.path.join(os.path.dirname(__file__), 'lib', 'libmbtable_sdk_shared.so')
lib = ctypes.CDLL(lib_path)


lib.MbTableOpen.argtypes = [ctypes.c_char_p]
lib.MbTableOpen.restype = ctypes.c_void_p

lib.MbTableClose.argtypes = [ctypes.c_void_p]
lib.MbTableClose.restype = None

lib.MbTableRead.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
lib.MbTableRead.restype = ctypes.POINTER(ByteArray)

lib.MbTableGetMetaData.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
lib.MbTableGetMetaData.restype = ctypes.POINTER(ByteArray)

lib.MbTableGetEntryCount.argtypes = [ctypes.c_void_p]
lib.MbTableGetEntryCount.restype = ctypes.c_int32

lib.MbTableCreateIterator.argtypes = [ctypes.c_void_p, ctypes.c_uint32]
lib.MbTableCreateIterator.restype = ctypes.c_void_p

lib.IteratorGetRecord.argtypes = [ctypes.c_void_p]
lib.IteratorGetRecord.restype = ctypes.POINTER(ByteArray)

lib.IteratorHasNext.argtypes = [ctypes.c_void_p]
lib.IteratorHasNext.restype = ctypes.c_bool

lib.IteratorNext.argtypes = [ctypes.c_void_p]
lib.IteratorNext.restype = None

lib.IteratorDelete.argtypes = [ctypes.c_void_p]
lib.IteratorDelete.restype = None

lib.FreeByteArray.argtypes = [ctypes.POINTER(ByteArray)]
lib.FreeByteArray.restype = None

class MbTable:
    def __init__(self, path: str, max_retry=5):
        self.path = path
        self.max_retry = max_retry
        
    def get_handle(self):
        self.handle = lib.MbTableOpen(self.path.encode('utf-8'))
        retry = 0
        while not self.handle and retry < self.max_retry:
            retry += 1
            print('Failed to open mbtable, retrying in 5 second')
            import time
            import random
            time.sleep(random.randint(1, 5))
            self.handle = lib.MbTableOpen(self.path.encode('utf-8'))
        if not self.handle:
            raise Exception('Failed to open mbtable after {} retries'.format(self.max_retry))

        return self.handle

    def read(self, index):
        '''
        only use this method if you seek few non-sequential records, otherwise use iterator
        '''
        self.get_handle()
        value_bytes = lib.MbTableRead(self.handle, index).contents
        value = ctypes.string_at(value_bytes.data, value_bytes.length)
        lib.FreeByteArray(value_bytes)
        self.close()
        return value

    def get_metadata(self, metadata_key):
        self.get_handle()
        key_bytes = metadata_key.encode('utf-8')
        value_bytes = lib.MbTableGetMetaData(self.handle, key_bytes).contents
        value = ctypes.string_at(value_bytes.data, value_bytes.length)
        lib.FreeByteArray(value_bytes)
        self.close()
        return value

    def get_entry_count(self):
        self.get_handle()
        count = lib.MbTableGetEntryCount(self.handle)
        self.close()
        return count

    def close(self):
        if self.handle:
            lib.MbTableClose(self.handle)
            self.handle = None

class MbTableIterator:
    def __init__(self, path, start_index=0, max_iterations=None):
        self.path = path
        self.start_index = start_index
        self.max_iterations = max_iterations
        self.current_iteration = 0
        self.iterator = None
        self.mbtable = None

    def __enter__(self):
        self.mbtable = MbTable(self.path)
        self.iterator = lib.MbTableCreateIterator(self.mbtable.get_handle(), self.start_index)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.iterator:
            lib.IteratorDelete(self.iterator)
        if self.mbtable:
            self.mbtable.close()

    def __iter__(self):
        return self

    def __next__(self):
        if self.max_iterations is not None and self.current_iteration >= self.max_iterations:
            raise StopIteration
        if lib.IteratorHasNext(self.iterator):
            self.current_iteration += 1
            value = self.get_record()
            lib.IteratorNext(self.iterator)
            return value
        else:
            raise StopIteration

    def next(self):
        # Python 2的兼容性方法
        return self.__next__()
    
    def get_record(self):
        ret = lib.IteratorGetRecord(self.iterator).contents
        record = ctypes.string_at(ret.data, ret.length)
        lib.FreeByteArray(ret)
        return record


if __name__ == '__main__':
    mbt_path = 'test/data/test.mbt'
    mbtable = MbTable(mbt_path)
    print(mbtable.get_entry_count())
    print(mbtable.read(1))
    print(mbtable.get_metadata('meta_key'))
    with MbTableIterator(mbt_path) as it:
        for record in it:
            print(record)


import itertools
import math
import operator
import random
import threading
import time
from typing import List
import torch
from modelbest_sdk.dataset.cache import AllCached, CacheFull, CacheNotFilled, PrefetchCache
from modelbest_sdk.dataset.range import Range
from modelbest_sdk.dataset.thrift_wrapper.base_doc import BaseDoc
from modelbest_sdk.dataset.thrift_wrapper.dataset_checkpoint import *
from modelbest_sdk.dataset.thrift_wrapper.dataset_context import DatasetContext
from modelbest_sdk.dataset.utils import get_worker_info
from modelbest_sdk.file_format.mbtable import MbTable, MbTableIterator
from modelbest_sdk.file_format.mbtable_partition import MbTablePartition, MbTablePartitionIterator


class MBTableIterableDataset(torch.utils.data.IterableDataset):
    def __init__(
        self, 
        context: DatasetContext, 
        path, 
        max_epoch,
        prefetch_chunk_cnt=20, 
        chunk_size=1024):
        
        self.context = context
        self.path = path
        self.max_epoch = max_epoch
        
        self.is_dir = os.path.isdir(path)
        if self.is_dir:
            self.mbtable_partition = MbTablePartition(path)
            self._length = self.mbtable_partition.get_total_count()
        
        else:
            self.mbtable = MbTable(path)
            self._length = self.mbtable.get_entry_count()
        
        self.prefetch_chunk_cnt = prefetch_chunk_cnt
        self.chunk_size, self.num_chunks = self.safe_chunk_size(chunk_size)
        
        self.cache = None
        self.exhausted = False
        self.used = Used()
        
    def __len__(self):
        return self._length
    
    def get_chunk_data(self, chunk):
        assert isinstance(chunk, Chunk)
        start_index = chunk.start
        max_iter = chunk.stop - chunk.start
        ret = []
        print(f"start_index: {start_index}, max_iter: {max_iter}")
        if self.is_dir:
            with MbTablePartitionIterator(self.mbtable_partition, start_index, max_iter) as iter:
                for record in iter:
                    ret.append(record)
        else:
            with MbTableIterator(self.mbtable, start_index, max_iter) as iter:
                for record in iter:
                    ret.append(record)
        return ret
    
    def __iter__(self):
        self.init_epoch()
        prefetch_thread = threading.Thread(target=self.prefetch)
        prefetch_thread.start()
        while True:
            try:
                chunk_data = self.cache.get()
                for data in chunk_data:
                    yield data
            except CacheNotFilled:
                time.sleep(0.1)
            except StopIteration:
                self.exhausted = True
                return
    
    def init_epoch(self):
        if self.used.epoch >= self.max_epoch:
            self.exhausted = True
        self.cache = PrefetchCache(size=self.prefetch_chunk_cnt)
        chunks = itertools.chain.from_iterable(
            (self.random_chunks(epoch) for epoch in range(self.used.epoch, self.max_epoch))
        )
        chunks = itertools.filterfalse(
            lambda chunk: chunk in self.used.done.get(chunk.epoch, set()), chunks
        )
        self.cache.submit(chunks)

    def safe_chunk_size(self, chunk_size):
        total_workers = self.context.num_workers * self.context.world_size
        if len(self) < total_workers:
            raise ValueError(
                f"more concurrent loaders ({total_workers}) than " f"data entries ({len(self)}) in '{self.path}'"
            )
        num_chunks = math.ceil(len(self) / chunk_size)
        if num_chunks <= total_workers:
            chunk_size = math.ceil(len(self) / total_workers)
            num_chunks = math.ceil(len(self) / chunk_size)
        return chunk_size, num_chunks

    def random_chunks(self, epoch):
        world_size = self.context.world_size
        rank = self.context.rank
        num_workers, worker_id = get_worker_info()
        # we only iteratre through start ids as they uniquely mark each slice
        r = Range(0, len(self), self.chunk_size)
        # split index among multi-gpu workers
        r = r.subrange(split=rank, nsplits=world_size)
        # split index among multi-process dataloader workers
        r = r.subrange(split=worker_id, nsplits=num_workers)
        # obtain random chunks
        chunks = (Chunk(epoch, st, min(st + self.chunk_size, len(self))) for st in r.random_iterate())
        return chunks

    def prefetch(self):
        while True:
            try:
                chunk = self.cache.pull_task()
                chunk_data = self.get_chunk_data(chunk)
                unused_chunk_data = [
                    dict(chunk=chunk, index=i, doc=BaseDoc.deserialize(data))
                    for i, data in zip(range(chunk.start, chunk.stop), chunk_data)
                    if (data is not None) and (i not in self.used.active.get(chunk, set()))
                ]
                self.cache.put_result(unused_chunk_data)
            except CacheFull:
                time.sleep(0.1)
            except AllCached:
                break

    def checkpoint(self):
        return DatasetCheckpoint(
            dataset_info=DatasetInfo(
                path=self.path, 
                max_epoch=self.max_epoch),
            used=self.used, 
            chunk_size=self.chunk_size, 
            num_chunks=self.num_chunks
            )
    
    def load_checkpoint(self, checkpoint: DatasetCheckpoint):
        if len(checkpoint.used.active) == 0 and len(checkpoint.used.done) == 0:
            self.used.epoch = 0
        else:
            epochs_not_exhausted = set()
            for chunk in checkpoint.used.active.keys():
                epochs_not_exhausted.add(chunk.epoch)
            epoch = 0
            for epoch in sorted(list(checkpoint.used.done.keys())):
                if len(checkpoint.used.done[epoch]) == checkpoint.num_chunks:
                    del checkpoint.used.done[epoch]
                else:
                    epochs_not_exhausted.add(epoch)
                    break
        epochs_not_exhausted.add(epoch + 1)
        checkpoint.used.epoch = min(epochs_not_exhausted, default=0)
        self.used = checkpoint.used        
    
    def update(self, entries: Dict[Chunk, List[int]]):
        for chunk, indexes in entries.items():
            self.used.active.setdefault(chunk, set()).update(set(indexes))
            
            if len(self.used.active[chunk]) == chunk.stop - chunk.start:
                self.used.done.setdefault(chunk.epoch, set()).add(chunk)
                del self.used.active[chunk]
            
            for epoch in range(chunk.epoch - 1, 0, -1):
                cur_chunk = Chunk(epoch, chunk.start, chunk.stop)
                if cur_chunk in self.used.active:
                    del self.used.active[cur_chunk]
                    self.used.done.setdefault(epoch, set()).add(cur_chunk)
                else:
                    break
    

if __name__ == '__main__':
    context = DatasetContext(world_size=1, rank=0, num_workers=1)
    dataset = MBTableIterableDataset(
        context=context, 
        path="/home/emr-user/modelbest_sdk/test/base_doc_simple",
        max_epoch=3)
    i = 0
    for data in dataset:
        entries = {data['chunk']: [data['index']]}
        dataset.update(entries=entries)
        i += 1
        if i == 29:
            checkpoint = dataset.checkpoint()
            break
    
    print(checkpoint)
    dataset_next = MBTableIterableDataset(
        context=context, 
        path="/home/emr-user/modelbest_sdk/test/base_doc_simple",
        max_epoch=3)
    dataset_next.load_checkpoint(checkpoint)
    for data in dataset_next:
        entries = {data['chunk']: [data['index']]}
        dataset_next.update(entries=entries)
        i += 1
        print(data)
        if i == 59:
            break
    checkpoint = dataset_next.checkpoint()
    print(checkpoint)


from setuptools import setup, find_packages

setup(
    name='modelbest_sdk',
    version='0.1',
    author='HankyZhao',
    author_email='zhq980115@gmail.com',
    description='Everything about modelbest data include data format mbtable, dataset, dataloader, and tools',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url='https://codeup.aliyun.com/64ddb0a87f62ff9b3d23ca15/modelbest_sdk',
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
)



import ctypes
import os

from modelbest_sdk.file_format.mbtable import ByteArray

lib_path = os.path.normpath(os.path.join(os.path.dirname(__file__), 'lib', 'libmbtable_sdk_shared.so'))
lib = ctypes.CDLL(lib_path)

lib.MbTableBuilderCreate.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
lib.MbTableBuilderCreate.restype = ctypes.c_void_p

lib.MbTableBuilderWrite.argtypes = [ctypes.c_void_p, ctypes.POINTER(ByteArray)]
lib.MbTableBuilderWrite.restype = ctypes.c_int

lib.MbTableBuilderAddMetaData.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]
lib.MbTableBuilderAddMetaData.restype = None

lib.MbTableBuilderBuild.argtypes = [ctypes.c_void_p]
lib.MbTableBuilderBuild.restype = None

def create_byte_array(data):
    return ByteArray(ctypes.cast(ctypes.c_char_p(data), ctypes.POINTER(ctypes.c_char)), len(data))

class MbTableBuilder:
    def __init__(self, path: str, codec: str = 'kZlib'):
        """
        Initializes an instance of the class, setting up the underlying MbTable builder with the specified path and compression codec.

        The `codec` parameter specifies the compression method to be used for the MbTable. It must be a string that matches one of the predefined compression codec names. The available codecs are as follows:
        - 'kLzo': LZO compression.
        - 'kZlib': zlib compression.
        - 'kUnCompress': No compression.
        - 'kGzip': Gzip compression.
        - 'kSnappy': Snappy compression.
        - 'kUnknown': Represents an unknown or unsupported compression codec. It's recommended to use one of the supported codecs for compatibility.

        Args:
            path (str): The file system path where the MbTable will be created.
            codec (str, optional): The name of the compression codec to use. Defaults to 'kZlib'.

        Note:
            The path and codec arguments are encoded in UTF-8 before being passed to the underlying `CreateMbTableBuilder` function of the `lib` library.
        """
        self.builder = lib.MbTableBuilderCreate(path.encode('utf-8'), codec.encode('utf-8'))

    def write(self, record):
        if isinstance(record, str):
            record = record.encode('utf-8')
        record_ba = create_byte_array(record)
        ret = lib.MbTableBuilderWrite(self.builder, ctypes.byref(record_ba))
        if ret == -1:
            raise Exception('Maximum of records 2^30 reached')
            

    def add_metadata(self, key: str, value: str):
        lib.MbTableBuilderAddMetaData(self.builder, key.encode('utf-8'), value.encode('utf-8'))

    def flush(self):
        lib.MbTableBuilderBuild(self.builder)

if __name__ == '__main__':
    builder = MbTableBuilder('test/data/test.mbt', 'kZlib')
    builder.write('value1')
    builder.write('value2')
    builder.write('value3')
    builder.add_metadata('meta1', 'value1')
    builder.add_metadata('meta2', 'value2')
    builder.flush()



import logging
import thriftpy2
from thriftpy2.utils import deserialize, serialize
import os

from modelbest_sdk.dataset.thrift_wrapper.utils import Utils

proto_dir = os.path.join(os.path.dirname(__file__), "../..", "proto")

doc_thrift = thriftpy2.load(os.path.join(proto_dir, "traindoc.thrift"), module_name="doc_thrift")
logger = logging.getLogger(__name__)


class BaseDoc:
    def __init__(self, token_ids, mask=None, docid=None, tag=None, token=None, tokenizer_version=None, reserved_col=None):
        self.token_ids = token_ids
        self.mask = mask
        self.docid = docid
        self.tag = tag
        self.token = token
        self.tokenizer_version = tokenizer_version
        self.reserved_col = reserved_col
    
    @staticmethod
    def deserialize(bin):
        return BaseDoc.from_thrift(deserialize(doc_thrift.BaseDoc(), bin))
    
    def to_thrift(self):
        return doc_thrift.BaseDoc(
            token_ids=self.token_ids,
            mask=self.mask,
            docid=self.docid,
            tag=self.tag,
            token=self.token,
            tokenizer_version=self.tokenizer_version,
            reserved_col=self.reserved_col
        )
    
    @staticmethod
    def from_thrift(thrift_base_doc):
        return BaseDoc(
            token_ids=thrift_base_doc.token_ids,
            mask=thrift_base_doc.mask,
            docid=thrift_base_doc.docid,
            tag=thrift_base_doc.tag,
            token=thrift_base_doc.token,
            tokenizer_version=thrift_base_doc.tokenizer_version,
            reserved_col=thrift_base_doc.reserved_col
        )
        
    def __repr__(self) -> str:
        return f"BaseDoc(token_ids={self.token_ids}, mask={self.mask}, docid={self.docid}, tag={self.tag}, token={self.token}, tokenizer_version={self.tokenizer_version}, reserved_col={self.reserved_col})"
    


import logging
from typing import Dict, List, Set
import thriftpy2
from thriftpy2.utils import deserialize, serialize
import os

from modelbest_sdk.dataset.thrift_wrapper.utils import Utils

proto_dir = os.path.join(os.path.dirname(__file__), "../..", "proto")

dc_thrift = thriftpy2.load(os.path.join(proto_dir, "dataset_checkpoint.thrift"), module_name="checkpoint_thrift")
dc_thrift.Chunk.__hash__ = lambda self: hash((self.epoch, self.start, self.stop))

logger = logging.getLogger(__name__)


class DatasetCheckpointList:
    def __init__(self, checkpoint_list, world_size=None, tp_size=None):
        self.checkpoint_list = checkpoint_list
        self.world_size = world_size
        self.tp_size = tp_size
        
    @staticmethod
    def load_from_file(path):
        return DatasetCheckpointList.from_thrift(deserialize(dc_thrift.DatasetCheckpointList(), Utils.load_from_file(path)))

    def save_to_file(self, path):
        Utils.save_to_file(path, serialize(self.to_thrift()))

    def to_thrift(self):
        thrift_checkpoints = [checkpoint.to_thrift() for checkpoint in self.checkpoint_list]
        return dc_thrift.DatasetCheckpointList(
            checkpoint_list=thrift_checkpoints,
            world_size=self.world_size,
            tp_size=self.tp_size
        )

    @staticmethod
    def from_thrift(thrift_checkpoint_list):
        checkpoint = [DatasetCheckpoint.from_thrift(checkpoint) for checkpoint in thrift_checkpoint_list.checkpoint_list]
        return DatasetCheckpointList(
            checkpoint_list=checkpoint,
            world_size=thrift_checkpoint_list.world_size,
            tp_size=thrift_checkpoint_list.tp_size
        )
        
    def __repr__(self) -> str:
        return f"DatasetCheckpointList(checkpoint_list={self.checkpoint_list}, world_size={self.world_size}, tp_size={self.tp_size})"
        


class DatasetCheckpoint:
    def __init__(self, dataset_info, used=None, chunk_size=None, num_chunks=None):
        self.dataset_info = dataset_info
        self.used: Used = used
        self.chunk_size = chunk_size
        self.num_chunks = num_chunks

    def to_thrift(self):
        thrift_dataset_info = self.dataset_info.to_thrift() if self.dataset_info else None
        thrift_used = self.used.to_thrift() if self.used else None
        return dc_thrift.DatasetCheckpoint(
            dataset_info=thrift_dataset_info,
            used=thrift_used,
            chunk_size=self.chunk_size,
            num_chunks=self.num_chunks
        )

    @staticmethod
    def from_thrift(thrift_checkpoint):
        dataset_info = DatasetInfo.from_thrift(thrift_checkpoint.dataset_info) if thrift_checkpoint.dataset_info else None
        used_ = Used.from_thrift(thrift_checkpoint.used) if thrift_checkpoint.used else None
        return DatasetCheckpoint(
            dataset_info=dataset_info,
            used=used_,
            chunk_size=thrift_checkpoint.chunk_size,
            num_chunks=thrift_checkpoint.num_chunks
        )
        
    def __repr__(self) -> str:
        return f"DatasetCheckpoint(dataset_info={self.dataset_info}, used={self.used}, chunk_size={self.chunk_size}, num_chunks={self.num_chunks})"


class DatasetInfoList:
    def __init__(self, dataset_info_list=None):
        self.dataset_info_list: List[DatasetInfo] = dataset_info_list if dataset_info_list is not None else []

    @staticmethod
    def load_from_file(path):
        return DatasetInfoList.from_thrift(deserialize(dc_thrift.DatasetInfoList(), Utils.load_from_file(path)))

    def save_to_file(self, path):
        Utils.save_to_file(path, serialize(self.to_thrift()))

    def to_thrift(self):
        return dc_thrift.DatasetInfoList(
            dataset_info_list=[info.to_thrift() for info in self.dataset_info_list]
        )

    @staticmethod
    def from_thrift(thrift_dataset_info_list):
        dataset_info_list = [DatasetInfo.from_thrift(info) for info in thrift_dataset_info_list.dataset_info_list]
        return DatasetInfoList(dataset_info_list=dataset_info_list)

    def __repr__(self) -> str:
        return f"DatasetInfoList(dataset_info_list={self.dataset_info_list})"

class DatasetInfo:
    def __init__(self, path, weight=None, max_epoch=None):
        self.path = path
        self.weight = weight
        self.max_epoch = max_epoch

    def to_thrift(self):
        return dc_thrift.DatasetInfo(
            path=self.path,
            weight=self.weight,
            max_epoch=self.max_epoch
        )

    @staticmethod
    def from_thrift(thrift_dataset_info):
        return DatasetInfo(
            path=thrift_dataset_info.path,
            weight=thrift_dataset_info.weight,
            max_epoch=thrift_dataset_info.max_epoch
        )
        
    def __repr__(self) -> str:
        return f"DatasetInfo(path={self.path}, weight={self.weight}, max_epoch={self.max_epoch})"


class Used:
    def __init__(self):
        self.active: Dict[Chunk, Set[int]] = {}
        self.done: Dict[int, Set[Chunk]] = {}
        self.epoch: int = 0

    def to_thrift(self):
        thrift_active = {chunk.to_thrift(): index_set for chunk, index_set in self.active.items()}
        thrift_done = {epoch: {chunk.to_thrift() for chunk in chunk_set} for epoch, chunk_set in self.done.items()}
        return dc_thrift.Used(active=thrift_active, done=thrift_done, epoch=self.epoch)

    @staticmethod
    def from_thrift(thrift_used):
        used = Used()
        used.epoch = thrift_used.epoch
        for thrift_chunk, index_set in thrift_used.active.items():
            used.active[Chunk.from_thrift(thrift_chunk)] = set(index_set)
        for epoch, thrift_chunk_set in thrift_used.done.items():
            used.done[epoch] = {Chunk.from_thrift(chunk) for chunk in thrift_chunk_set}
        return used
    
    def __repr__(self) -> str:
        return f"Used(active={self.active}, done={self.done}, epoch={self.epoch})"


class Chunk:
    def __init__(self, epoch, start, stop):
        self.epoch = epoch
        self.start = start
        self.stop = stop

    def to_thrift(self):
        return dc_thrift.Chunk(epoch=self.epoch, start=self.start, stop=self.stop)

    @staticmethod
    def from_thrift(thrift_chunk):
        return Chunk(epoch=thrift_chunk.epoch, start=thrift_chunk.start, stop=thrift_chunk.stop)
    
    def __repr__(self) -> str:
        return f"Chunk(epoch={self.epoch}, start={self.start}, stop={self.stop})"

    def __hash__(self) -> int:
        return hash((self.epoch, self.start, self.stop))
    
    def __eq__(self, other) -> bool:
        return self.epoch == other.epoch and self.start == other.start and self.stop == other.stop